###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import os
from io import StringIO
from subprocess import run as command, PIPE, DEVNULL
from types import TracebackType

from paramiko import SFTPClient, SFTPAttributes, SSHClient, rsakey

from everysk.config import settings
from everysk.core.datetime import Date, DateTime
from everysk.core.object import BaseObject


###############################################################################
#   KnownHosts Class Implementation
###############################################################################
class KnownHosts(BaseObject):
    content: dict[str, str] = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.load()

    @property
    def filename(self) -> str:
        """
        Get the known_hosts file full path.
        """
        return os.path.join(settings.EVERYSK_SFTP_DIR, 'known_hosts')

    def _verify_file_exist(self) -> bool:
        # If the file already exists we stop here
        if os.path.exists(self.filename):
            return True

        # If the directory does not exist we need to create it
        dir = os.path.dirname(self.filename) # pylint: disable=redefined-builtin
        if not os.path.exists(dir):
            os.makedirs(dir)

        # If the file does not exist, we create an empty one
        # https://stackoverflow.com/a/12654798
        open(self.filename, 'w', encoding='utf-8').close() # pylint: disable=consider-using-with
        return True

    def add(self, hostname: str) -> None:
        """
        Add the hostname to the known_hosts in the cache and local file.

        Args:
            hostname (str): The hostname to add to the known_hosts. Example: 'files.example.com'
        """
        # Use the ssh-keyscan to get the key of the hostname
        result = command(['ssh-keyscan', hostname], stdout=PIPE, check=False, stderr=DEVNULL)
        # Add the key to the known_hosts locally
        self.content[hostname] = result.stdout.decode('utf-8')
        # Save the known_hosts to the local file
        self.write()

    def check(self, hostname: str) -> bool:
        """
        Check if the hostname is already in the known_hosts.

        Args:
            hostname (str): The hostname to check in the known_hosts. Example: 'files.example.com'
        """
        return hostname in self.content

    def delete(self, hostname: str) -> None:
        """
        Delete the hostname from the known_hosts in the cache and local file.

        Args:
            hostname (str): The hostname to delete from the known_hosts. Example: 'files.example.com'
        """
        if hostname in self.content:
            del self.content[hostname]
            self.write()

    def load(self) -> None:
        """
        Load the known_hosts from the local file $HOME/.ssh/known_hosts.
        """
        if self._verify_file_exist():
            with open(self.filename, 'r', encoding='utf-8') as fd:
                self.content = {line.split(' ')[0]: line for line in fd.readlines()}

    def write(self) -> None:
        """
        Write the known_hosts to the local file $HOME/.ssh/known_hosts for future use.
        """
        if self._verify_file_exist():
            with open(self.filename, 'w', encoding='utf-8') as fd:
                fd.writelines(self.content.values())
                # Ensure the file is written to disk
                fd.flush()
                os.fsync(fd.fileno())


###############################################################################
#   SFTP Class Implementation
###############################################################################
class SFTP(BaseObject):
    """
    SFTP class to connect to the SFTP server.
    We could use the context manager to automatically close the connection when the object is deleted/destroyed.

    Args:
        compress (bool): If the connection will transfer compressed data. Defaults to True.
        date (Date, DateTime): The date/datetime used to parse the name. Defaults to today.
        hostname (str): The hostname of the SFTP server. Defaults to None.
        password (str): The password of the SFTP server. Defaults to None.
        private_key (str): The private key of the SFTP server. Defaults to None.
        passphrase (str): The passphrase of the private key. Defaults to None.
        port (int): The port of the SFTP server. Defaults to 22.
        timeout (int): The timeout of the SFTP connection. Defaults to 60.
        username (str): The username of the SFTP server. Defaults to None.

    Example:

        >>> from everysk.core.sftp import SFTP
        >>> with SFTP(username='', password='', hostname='') as sftp:
        ...     filename = sftp.search_by_last_modification_time(path='/dir', prefix='file_')
        >>> print(filename)
        /dir/2024/11/13/file_11.12.2024.csv
    """
    ## Private attributes
    _client: SFTPClient = None

    ## Public attributes
    compress: bool = True
    date: Date | DateTime = None
    hostname: str = None
    password: str = None
    private_key: str = None
    passphrase: str = None
    port: int = 22
    timeout: int = 60
    username: str = None

    @property
    def client(self) -> SFTPClient:
        """
        Get the SFTP client to connect to the SFTP server.
        """
        if self._client is None:
            self._client = self.get_sftp_client(
                hostname=self.hostname,
                port=self.port,
                username=self.username,
                password=self.password,
                private_key=self.private_key,
                passphrase=self.passphrase,
                compress=self.compress,
                timeout=self.timeout
            )

        return self._client

    ## Private methods
    def __init__(
        self,
        hostname: str,
        username: str,
        password: str = None,
        port: int = 22,
        private_key: str = None,
        passphrase: str = None,
        date: Date | DateTime = None,
        compress: bool = True,
        timeout: int = 60,
        **kwargs
    ):
        """
        Constructor to initialize the SFTP connection.

        Args:
            hostname (str, optional): The hostname of the SFTP server. Defaults to None.
            username (str, optional): The username of the SFTP server. Defaults to None.
            password (str, optional): The password of the SFTP server. Defaults to None.
            port (int, optional): The port of the SFTP server. Defaults to 22.
            private_key (str, optional): The private key of the SFTP server. Defaults to None.
            passphrase (str, optional): The passphrase of the private key. Defaults to None.
            date (Date, DateTime, optional): The date/datetime used to parse the name. Defaults to today.
            compress (bool, optional): If the connection will transfer compressed data. Defaults to True.
            timeout (int, optional): The timeout of the SFTP connection. Defaults to 60.
        """
        if date is None:
            date = Date.today()

        super().__init__(
            compress=compress,
            date=date,
            hostname=hostname,
            password=password,
            port=port,
            private_key=private_key,
            passphrase=passphrase,
            timeout=timeout,
            username=username,
            **kwargs
        )

    def __del__(self):
        """
        Destructor to close the SFTP connection when the object is deleted/destroyed.
        """
        try:
            if self._client is not None:
                # Close the SFTP connection when the object is deleted/destroyed
                self._client.close()
        except Exception: # pylint: disable=broad-except
            pass

    def __enter__(self):
        """
        Enter the context manager to return the object itself.
        """
        return self

    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None):
        """
        https://docs.python.org/3/library/stdtypes.html#contextmanager.__exit__

        Returns:
            bool | None: If return is False any exception will be raised.
        """
        try:
            if self._client is not None:
                # Close the SFTP connection
                self._client.close()
                # Reset the SFTP client
                self._client = None
        except Exception: # pylint: disable=broad-except
            pass

    def sort(self, lst: list, attr: str, reverse: bool = False) -> list:
        """
        Sort the list of objects by the attribute with order by asc or desc.
        If the attribute is not found, the list will be returned as is.
        If reverse is True, the list will be sorted in descending order.

        Args:
            lst (list): The list of objects to sort.
            attr (str): The name of the attribute to sort.
            reverse (bool, optional): The final order of the list. Defaults to False.
        """
        return sorted(lst, key=lambda obj: getattr(obj, attr), reverse=reverse)

    ## Public methods
    def get_sftp_client(self, hostname: str, port: int, username: str, password: str = None, compress: bool = None, timeout: int = None, private_key: str = None, passphrase: str = None) -> SFTPClient:
        """
        Connect to the SFTP server and return the SFTP client.

        Args:
            hostname (str): The hostname of the SFTP server.
            port (int): The port of the SFTP server.
            username (str): The username of the SFTP server.
            password (str): The password of the SFTP server.
            compress (bool): If the connection will transfer compressed data.
            timeout (int): The timeout of the SFTP connection.
            private_key (str): The private key of the SFTP server.
            passphrase (str): The passphrase of the private key.
        """
        ssh = SSHClient()
        # On this point we create the file if it does not exists
        known_hosts = KnownHosts()
        # Check if the hostname is already known
        if not known_hosts.check(hostname):
            # Add the hostname to the known_hosts and write to the local file
            known_hosts.add(hostname)

        params = {
            'hostname': hostname,
            'port': port,
            'username': username,
            'compress': compress,
            'timeout': timeout,
        }

        # check if we have the password
        if password is not None:
            params['password'] = password
        elif private_key is not None:
            key_file = StringIO(private_key)
            params['pkey'] = rsakey.RSAKey.from_private_key(key_file)
            params['passphrase'] = passphrase

        # Load the known_hosts file
        # This was placed before the connection to have time for Google sync the file creation
        # https://everysk.atlassian.net/browse/COD-10872
        ssh.load_system_host_keys(filename=known_hosts.filename)

        ssh.connect(**params)
        return ssh.open_sftp()

    def get_file(self, filename: str) -> bytes | None:
        """
        Get the file content from the SFTP server.
        If the file is not found, return None.
        If the filename has a date format, it will be parsed with the date attribute.
        Example: '/dir/%Y/file_%Y.csv' -> '/dir/2024/file_2024.csv'

        Args:
            filename (str): The filename with the path to get the content. Example: '/dir/2024/file_2024.csv'
        """
        if '%' in filename:
            filename = self.parse_date(filename, self.date)

        try:
            with self.client.open(filename, 'rb') as fd:
                return fd.read()
        except IOError:
            # If the file is not found, return None
            pass

        return None

    def save_file(self, filename: str, content: bytes) -> None:
        """
        Save the file content to the SFTP server.
        If the filename has a date format, it will be parsed with the date attribute.
        Example: '/dir/%Y/file_%Y.csv' -> '/dir/2024/file_2024.csv'

        Args:
            filename (str): The filename with the path to save the content. Example: '/dir/2024/file_2024.csv'
            content (bytes): The content of the file to save.
        """
        if '%' in filename:
            filename = self.parse_date(filename, self.date)

        # Split the path to create the directories if not exists
        # the first element is empty, so we need to remove it
        dirs = os.path.dirname(filename).split('/')[1:]
        path = ''
        for dir in dirs: # pylint: disable=redefined-builtin
            path = f'{path}/{dir}'
            try:
                self.client.mkdir(path)
            except IOError:
                pass

        with self.client.open(filename, 'wb') as fd:
            fd.write(content)

    def search_by_last_modification_time(self, path: str, prefix: str) -> str | None:
        """
        Search the file by the last modification time with the prefix in the path recursively.
        If the file is not found, return None.
        If the path or prefix has a date format, it will be parsed with the date attribute.
        Example: '/dir/%Y' -> '/dir/2024' or 'file_%m.%d.%Y' -> 'file_11.30.2024'

        Args:
            path (str): The path to start search the file.
            prefix (str): The prefix of the file to search. Example: 'file_' or 'file_%m.%d.%Y'
        """
        if '%' in prefix:
            prefix = self.parse_date(prefix, self.date)
        if '%' in path:
            path = self.parse_date(path, self.date)

        objs: list[SFTPAttributes] = self.client.listdir_attr(path)
        objs = self.sort(objs, 'st_mtime', reverse=True)
        for file in objs:
            if file.filename.startswith(prefix):
                return f'{path}/{file.filename}'

            if file.longname.startswith('d'):
                result = self.search_by_last_modification_time(f'{path}/{file.filename}', prefix)
                if result:
                    return result

        return None

    def parse_date(self, name: str, date: Date) -> str:
        """
        Parse the date format in the name with the date attribute.
        Example: '/dir/%Y/file_%Y.csv' -> '/dir/2024/file_2024.csv'

        Args:
            name (str): The name with the date format to parse.
            date (Date): The date to parse the date format.
        """
        if '%' in name:
            index = name.find('%')
            date_format = name[index:index + 2]
            result = date.strftime(date_format)
            name = name.replace(date_format, result)
            if '%' in name:
                return self.parse_date(name, date)

        return name



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import re
from collections.abc import Iterator
from sys import maxsize as max_int
from types import UnionType
from typing import Any, Callable
from urllib.parse import urlparse

from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import ReadonlyError, FieldValueError
from everysk.core.object import BaseField, BaseDict
from everysk.utils import bool_convert


###############################################################################
#   Private Functions Implementation
###############################################################################
def _do_nothing(*args, **kwargs):
    # pylint: disable=unused-argument
    raise ReadonlyError('This field value cannot be changed.')

def _min_max_validate(min_value: Any, max_value: Any, value: Any, attr_name: str) -> None:
    """
    Helper function to validate if value is between min and max for fields.

    Args:
        min_value (Any): The min value to be checked.
        max_value (Any): The max value to be checked.
        value (Any): The value used to validate.
        attr_name (str): The name of the attribute to be used on errors.

    Raises:
        FieldValueError: If the value is not between min and max.

    Example:
        >>> _min_max_validate(0, 10, 5, "my_field")
        # No exception is raised

        >>> _min_max_validate(0, 10, 15, "my_field")
        # FieldValueError is raised with the message:
        # "The value '15' for field 'my_field' must be between 0 and 10."
    """
    if value is not None and value is not Undefined:
        min_check = False
        max_check = False

        # Check if min_value and max_value are callable
        min_value = min_value if not callable(min_value) else min_value()
        max_value = max_value if not callable(max_value) else max_value()

        if min_value is not None and min_value is not Undefined:
            min_check = value < min_value

        if max_value is not None and max_value is not Undefined:
            max_check = value > max_value

        if min_check or max_check:
            raise FieldValueError(
                f"The value '{value}' for field '{attr_name}' must be between {min_value} and {max_value}."
            )


###############################################################################
#   Field Class Implementation
###############################################################################
class Field(BaseField):
    attr_type: Any = Any
    choices: set = None

    def __new__(cls, *args, **kwargs) -> Any:
        """
        For the VSCode autocomplete works correctly, we need to say what is the
        type that the method __new__ returns
        """
        return super().__new__(cls)

    def __init__(
        self,
        default: Any = None,
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        super().__init__(attr_type=self.attr_type, default=default, required=required, readonly=readonly, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)
        if self.choices is None:
            self.choices = set()

        if self.choices and self.default and self._check_invalid_choices(self.default):
            raise FieldValueError(
                f"The default value '{self.default}' must be in this list {self.choices}."
            )

    def _check_invalid_choices(self, value: Any) -> bool:
        """
        Validates whether the given value is within the allowed choices.

        If `self.attr_type` is a list, tuple, or set, it checks whether all elements in `value`
        exist in `self.choices`. If `value` is a single item, it verifies its presence in `self.choices`.

        Args:
            value (Any): The value or collection of values to validate.

        Returns:
            bool: True if the value is invalid (not in `self.choices`), otherwise False.
        """
        raise_value_error = False
        if self.attr_type in {list, tuple, set}:
            if set(value).difference(self.choices):
                raise_value_error = True
        elif value not in self.choices:
            raise_value_error = True

        return raise_value_error

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        # Execute the normal validations
        super().validate(attr_name, value, attr_type)

        # Then we check if the value is in the choices
        if value is not None and value is not Undefined and self.choices and self._check_invalid_choices(value):
            raise FieldValueError(
                f"The value '{value}' for field '{attr_name}' must be in this list {self.choices}."
            )


###############################################################################
#   BoolField Class Implementation
###############################################################################
class BoolField(Field):
    attr_type: bool = bool

    def __new__(cls, *args, **kwargs) -> bool:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Converts the  given value to a boolean if possible using the 'convert_boolean' function.
        (Check the function's documentation for more information)

        Args:
            value (Boolean): The value to be converted. Can be of any type

        Returns:
            Any: The value converted to its boolean corresponding

        Example:
            >>> from everysk.core.fields import BoolField
            >>> bool_field = BoolField()
            >>> bool_field.clean_value("y")
            >>> True

            >>> bool_field.clean_value("n")
            >>> False

            >>> bool_field.clean_value("a")
            >>> ValueError: Invalid truth value 'a'
        """
        # https://docs.python.org/3/distutils/apiref.html#distutils.util.strtobool
        # The module distutils is deprecated, then we put the function code here
        if value is not None and value is not Undefined:
            value = bool_convert(value)

        return super().clean_value(value)


###############################################################################
#   ChoiceField Class Implementation
###############################################################################
class ChoiceField(Field):
    attr_type: str = str
    choices: list[str] = None

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: str,
        choices: list,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default=default, required=required, readonly=readonly, choices=choices, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

        # Choices always need to be a list
        if not self.choices:
            self.choices = []

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks if value is in list of choices.
        We discard the value Undefined.

        Args:
            attr_name (str): Name of the attribute being validated
            value (Any): The value up for validation against the choices available.
            attr_type (type | UnionType, optional): Type of the field we are trying to validate. Defaults to None.

        Returns:
            None: This method does not return any value. It either completes successfully or raises an exception

        Raises:
            FieldValueError: If value type doesn't match with the required type.

        Example:
            >>> from everysk.core.fields import ChoiceField
            >>> from everysk.core.exceptions import FieldValueError
            >>> field = ChoiceField(default="option1", choices=["option1", "option2", "option3"])
            >>> try:
            >>> ... field.validate("test_field", "option1")
            >>> ... print("Validation successful")
            >>> except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation successful

            >>> try:
            >>> ... field.validate("test_field", "option4")
            >>> ... print("Validation successful")
            >>> except Exception as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation error: The value 'option4' for field 'test_field' must be in this list ['option1', 'option2', 'option3'].
        """
        if value is not Undefined and value not in self.choices:
            raise FieldValueError(
                f"The value '{value}' for field '{attr_name}' must be in this list {self.choices}."
            )
        super().validate(attr_name, value, attr_type)


###############################################################################
#   DateField Class Implementation
###############################################################################
class DateField(Field):
    attr_type: Date = Date
    min_date: Date | Callable = None
    max_date: Date | Callable = None

    def __new__(cls, *args, **kwargs) -> Date:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_date: Date = None,
        max_date: Date = None,
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        super().__init__(default, min_date=min_date, max_date=max_date, required=required, readonly=readonly, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        This method simple converts Date strings to Date object in the Everysk format

        Args:
            value (Any): Desired date string to be converted

        Returns:
            A new Date object representing the converted date

        Raises:
            ValueError: If the string is not represented in either ISO format ("YYYY-MM-DD") or Everysk format ("YYYYMMDD").

        Example:
            >>> from everysk.core.fields import DateField
            >>> date_field = DateField()
            >>> date_field.clean_value("20140314")
            Date(2014, 3, 14)
            >>> date_field.clean_value("2014-03-14")
            Date(2014, 3, 14)
        """
        if isinstance(value, str):
            if '-' in value:
                value = Date.fromisoformat(value)
            else:
                # Everysk format
                value = Date.strptime(value, '%Y%m%d')

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks if value is greater than min and lower than max including both values.

        Args:
            attr_name (str): Name of the attribute used for error checking
            value (Any): Value used for validation
            attr_type (type | UnionType, optional): Type of the field we are trying to validate. Defaults to None.

        Returns:
            Either the value is between min_date and max_date

        Raises:
            FieldValueError: If the value is not between min_date and max_date

        Example:
            >>> from everysk.core.fields import DateField
            >>> from everysk.core.datetime import Date
            >>> date_field = DateField(min_date=Date(2023, 1, 1), max_date=Date(2023, 12, 31))

            >>> try:
            >>> ... date_field.validate("example_date", Date(2023, 6, 15))
            >>> ... print("June 15, 2023, is a valid date")
            >>> except Exception as e:
            >>> ... print(f"Validation error: {e}")
            >>> June 15, 2023, is a valid date

            >>> try:
            >>> ... date_field.validate("example_date", Date(2022, 12, 31))
            >>> ... print("December 31, 2022, is a valid date")
            >>> except Exception as e:
            >>> ... print(f"validation error: {e}")
            >>> Validation error: The value '2022-12-31' for field 'example_date' must be between 2023-01-01 and 2023-12-31.
        """
        _min_max_validate(self.min_date, self.max_date, value, attr_name)
        return super().validate(attr_name, value, attr_type)


###############################################################################
#   DateTimeField Class Implementation
###############################################################################
class DateTimeField(Field):
    attr_type:DateTime = DateTime
    min_date: DateTime | Callable = None
    max_date: DateTime | Callable = None
    force_time: str = None

    def __new__(cls, *args, **kwargs) -> DateTime:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_date: DateTime = None,
        max_date: DateTime = None,
        force_time: str = 'FIRST_MINUTE',
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        super().__init__(default, min_date=min_date, max_date=max_date, force_time=force_time, required=required, readonly=readonly, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Converts a DateTime string to a DateTime object using the 'fromisoformat' function.
        (check 'fromisoformat' function for more information)

        Args:
            value (Any): desired DateTime string to be converted

        Returns:
            Any: a new DateTime object formatted

        Raises:
            ValueError: if the input is not formatted in the ISO format

        Example:
            >>> from everysk.core.fields import DateTimeField
            >>> date_time_field = DateTimeField()

            >>> date_time_field.clean_value("2023-03-15")
            >>> DateTime(2023, 3, 15, 0, 0, tzinfo=zoneinfo.ZoneInfo(key='UTC'))

            >>> date_time_field.clean_value("2023-03-15T14:30:00")
            >>> DateTime(2023, 3, 15, 14, 30, tzinfo=zoneinfo.ZoneInfo(key='UTC'))

            >>> date_time_field.clean_value("14-03-2023")
            >>> ValueError: time data "14-03-2023" does not match format: "%Y-%m-%d"
        """
        # Convert DateTime strings to DateTime object
        if isinstance(value, str):
            if not ':' in value:
                value: DateTime = DateTime.fromisoformat(value).force_time(self.force_time)
            else:
                value: DateTime = DateTime.fromisoformat(value)
        elif Date.is_date(value):
            value: DateTime = DateTime.fromisoformat(value.isoformat())

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks is the value passed is in the correct range between min and max, as well as required, required_lazy, and attribute_type validations.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional):
                The expected type of the attribute. If provided, the value is checked to ensure it is of this type.
                Defaults to None.

        Raises:
            FieldValueError: if the value provided is not between min and max values

        Example:
            >>> from everysk.core.datetime import Date
            >>> from everysk.core.fields import DateTimeField
            >>> from everysk.core.datetime import DateTime
            >>> date_time_field = DateTimeField(min_date=DateTime(2023, 1, 1), max_date=DateTime(2023, 12, 31))

            >>> try:
            >>> ... date_time_field.validate("test_field", DateTime(2023, 6, 15))
            >>> ... print("Validation successful")
            >>> except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
        """
        _min_max_validate(self.min_date, self.max_date, value, attr_name)
        return super().validate(attr_name, value, attr_type)


###############################################################################
#   DictField Class Implementation
###############################################################################
class DictField(Field):
    # Field that supports a dictionary and BaseDict as a value.
    attr_type: dict | BaseDict = dict | BaseDict

    class ReadonlyDict(dict):
        __setitem__ = _do_nothing
        __delitem__ = _do_nothing
        pop = _do_nothing
        popitem = _do_nothing
        clear = _do_nothing
        update = _do_nothing
        setdefault = _do_nothing

    def __new__(cls, *args, **kwargs) -> dict | BaseDict:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, default: Any = None, required: bool = False, readonly: bool = False, required_lazy: bool = False, empty_is_none: bool = False, **kwargs) -> None:
        # When the field is readonly, we need to change the content to.
        if readonly and default:
            default = self.ReadonlyDict(default)

        super().__init__(default, required, readonly, required_lazy, empty_is_none, **kwargs)


###############################################################################
#   EmailField Class Implementation
###############################################################################
class EmailField(Field):
    attr_type: str = str

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Validates if the value is an e-mail address.
        To validate we check the existence of the '@' character and the length of the string.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional): The expected type of the attribute. If provided, the value is checked to ensure it is of this type. Defaults to None.

        Raises:
            FieldValueError: If the value is not an e-mail address.
        """
        # The maximum length of an email is 320 characters per RFC 3696 section 3,
        # and at least 3 digits a@b
        if value is not None and value is not Undefined and isinstance(value, str) and ('@' not in value or len(value) > 320 or len(value) < 3):
            raise FieldValueError(f'Key {attr_name} must be an e-mail.')

        try:
            # Get the cases where we could have more than one @ in the string
            user, domain = value.split('@') # pylint: disable=unused-variable
        except ValueError as error:
            raise FieldValueError(f'Key {attr_name} must be an e-mail.') from error

        return super().validate(attr_name, value, attr_type)


###############################################################################
#   FloatField Class Implementation
###############################################################################
class FloatField(Field):
    attr_type: float = float
    min_size: float = None
    max_size: float = None

    def __new__(cls, *args, **kwargs) -> float:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(self,
        default: Any = None,
        min_size: float = float('-inf'),
        max_size: float = float('inf'),
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        super().__init__(default, required, readonly, min_size=min_size, max_size=max_size, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Convert a float string to a float object

        Args:
            value (Any): Value desired for conversion to float

        Returns:
            Float: if value is able to be converted to float object

        Raises:
            ValueError: If value could not be converted to a float

        Example:
            >>> from everysk.core.fields import FloatField
            >>> float_field = FloatField()

            >>> float_field.clean_value("3.14")
            >>> 3.14

            >>> float_field.clean_value(123)
            >>> 123.0

            >>> float_field.clean_value('abc')
            >>> ValueError: could not convert string to float: 'abc'
        """
        # Convert Float strings to float object
        if isinstance(value, (int, str)):
            value = float(value)

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks if value is greater than min and lower than max including both values.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional): The expected type of the attribute. If provided, the value is checked to ensure it is of this type. Defaults to None.

        Raises:
            FieldValueError: If value is not between min and max.

        Example:
            >>> from everysk.core.fields import FloatField
            >>> from everysk.core.exceptions import FieldValueError
            >>> float_field = FloatField(min_size=0.0, max_size=100.0)

            >>> try:
            >>> ... float_field.validate("test_field", 50.0)
            >>> ... print("Validation successful: 50.0 is within the range.")
            >>>  except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation successful: 50.0 is within the range.

            >>> try:
            >>> ... float_field.validate("test_field", -10.0)
            >>> ... print("Validation successful: -10.0 is within the range.")
            >>>  except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation error: The value '-10.0' for field 'test_field' is not within the allowed range.
        """
        _min_max_validate(min_value=self.min_size, max_value=self.max_size, value=value, attr_name=attr_name)
        return super().validate(attr_name, value, attr_type)


###############################################################################
#   IntField Class Implementation
###############################################################################
class IntField(Field):
    attr_type: int = int
    min_size: int = None
    max_size: int = None

    def __new__(cls, *args, **kwargs) -> int:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(self,
        default: Any = None,
        min_size: int = -max_int,
        max_size: int = max_int,
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        super().__init__(default, required, readonly, min_size=min_size, max_size=max_size, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Converts a value to an integer if it can be converted

        Args:
            value (Any): Value to convert

        Returns:
            Int: value converted to an integer

        Raises:
            ValueError: if the string cannot be converted to an integer

        Example:
            >>> from everysk.core.fields import IntField
            >>> int_field = IntField()

            >>> int_field.clean_value("123")
            >>> 123

            >>> int_field.clean_value("abc")
            >>> ValueError: invalid literal for int() with base 10: 'abc'
        """
        # Convert Int strings to int object
        if isinstance(value, str):
            value = int(value)

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks if value is greater than min and lower than max including both values.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional): The expected type of the attribute. If provided, the value is checked to ensure it is of this type. Defaults to None.

        Raises:
            FieldValueError: If value is not between min_size and max.

        Example:
            >>> from everysk.core.fields import IntField
            >>> from everysk.core.exceptions import FieldValueError
            >>> int_field = IntField(min_size=1, max_size=10)

            >>> try:
            >>> ... int_field.validate("test_field", 5)
            >>> ... print("Validation successful")
            >>> except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation successful

            >>> try:
            >>> ... int_field.validate("test_field", -1)
            >>> ... print("validation successful")
            >>> except FieldValueError as e:
            >>> ...print(f"Validation error: {e}")
            >>> Validation error: The value '-1' for field 'test_field' must be between 1 and 10.
        """
        _min_max_validate(min_value=self.min_size, max_value=self.max_size, value=value, attr_name=attr_name)
        return super().validate(attr_name, value, attr_type)


###############################################################################
#   IteratorField Class Implementation
###############################################################################
class IteratorField(Field):
    attr_type: Iterator = Iterator

    def __new__(cls, *args, **kwargs) -> Iterator:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Cleans the given value before storing it.

        Args:
            value (Any): The value to be cleaned.

        Returns:
            Any: The cleaned value.

        """
        # Convert List/Str to iterators
        if isinstance(value, (list, str)):
            value = iter(value)

        return super().clean_value(value)


###############################################################################
#   ListField
###############################################################################
class ListField(Field):
    """
    https://stackoverflow.com/questions/855191/how-big-can-a-python-list-get#comment112727918_15739630
    so on 64 systems is maxsize / 8
    """
    attr_type: list = list
    min_size: int = None
    max_size: int = None
    separator: str = ','

    class ReadonlyList(list):
        __setitem__ = _do_nothing
        __delitem__ = _do_nothing
        append = _do_nothing
        clear = _do_nothing
        extend = _do_nothing
        insert = _do_nothing
        pop = _do_nothing
        remove = _do_nothing
        reverse = _do_nothing
        sort = _do_nothing

    def __new__(cls, *args, **kwargs) -> list:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_size: int = 0,
        max_size: int = max_int / 8,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        separator: str = ',',
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        if min_size < 0:
            raise FieldValueError('List min_size cloud not be a negative number.')

        # When the field is readonly, we need to change the content to.
        if readonly and (default is not None or default is not Undefined):
            default = self.ReadonlyList(default)

        super().__init__(
            default=default,
            min_size=min_size,
            max_size=max_size,
            readonly=readonly,
            required=required,
            required_lazy=required_lazy,
            separator=separator,
            empty_is_none=empty_is_none,
            **kwargs
        )

    def clean_value(self, value: Any) -> Any:
        """
        Clean the value before storing it.
        This method is used to ensure that the value is in the correct format before storing it.
        This method checks if the value is a string and converts it into a list if necessary.
        It then calls the parent class's clean_value method to perform additional cleaning.

        Args:
            value (Any): The value to be cleaned.

        Returns:
            Any: The cleaned value.

        Example:
            >>> field = Field()
            >>> field.clean_value("example")
            ['example']
        """
        if isinstance(value, str):
            # When we receive environment vars or HTTP query params they are always strings with ',' as separator
            value = [item.strip() for item in value.split(self.separator)]

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks if value is greater than min_size and lower than max_size including both values.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional): The expected type of the attribute. If provided, the value is checked to ensure it is of this type. Defaults to None.

        Raises:
            FieldValueError: If value is not list instance or if is not between min_size and max_size
        """
        if value is not None and value is not Undefined:
            if not isinstance(value, list):
                raise FieldValueError(f"The '{attr_name}' value must be a list.")
            if not self.min_size <= len(value) <= self.max_size:
                raise FieldValueError(
                    f"The attribute '{attr_name}' is not within the specified list range. min_size: {self.min_size} max_size: {self.max_size}"
                )
        super().validate(attr_name, value, attr_type)


###############################################################################
#   SetField Class Implementation
###############################################################################
class SetField(Field):
    attr_type: set = set
    min_size: int = None
    max_size: int = None

    class ReadonlySet(set):
        __setitem__ = _do_nothing
        __delitem__ = _do_nothing
        add = _do_nothing
        clear = _do_nothing
        discard = _do_nothing
        pop = _do_nothing
        remove = _do_nothing
        update = _do_nothing

    def __new__(cls, *args, **kwargs) -> set:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_size: int = 0,
        max_size: int = max_int / 8,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        if min_size < 0:
            raise FieldValueError('Set min_size cloud not be a negative number.')

        # When the field is readonly, we need to change the content to.
        if readonly and (default is not None or default is not Undefined):
            default = self.ReadonlySet(default)

        super().__init__(default=default, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Checks if value is greater than min_size and lower than max_size including both values.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional): The expected type of the attribute. If provided, the value is checked to ensure it is of this type. Defaults to None.

        Raises:
            FieldValueError: If value is not set instance.
            FieldValueError: If value is not between min_size and max_size.
        """
        if value is not None and value is not Undefined and isinstance(value, set):
            if not self.min_size <= len(value) <= self.max_size:
                raise FieldValueError(
                    f"The attribute '{attr_name}' is not within the specified set range. min_size: {self.min_size} max_size: {self.max_size}"
                )
        super().validate(attr_name, value, attr_type)


###############################################################################
#   StrField Class Implementation
###############################################################################
class StrField(Field):
    """
    Represents a string field in an entity.
    This field type is used to store string values and provides validation
    for string attributes based on specified constraints such as minimum and maximum size
    and regular expression pattern matching.

    Args:
        default (Any, optional): The default value for the field. Defaults to None.
        min_size (int, optional): The minimum allowed length for the string. Defaults to 0.
        max_size (int, optional): The maximum allowed length for the string. Defaults to the maximum integer value.
        regex (str, optional): The regular expression pattern to match against the string value. Defaults to None.
        readonly (bool, optional): Whether the field is read-only. Defaults to False.
        required (bool, optional): Whether the field is required. Defaults to False.
        required_lazy (bool, optional): Whether the field is lazily required. Defaults to False.
        empty_is_none (bool, optional): Whether an empty string should be treated as None. Defaults to False.

    Raises:
        FieldValueError: If the minimum size is negative
    """

    attr_type: str = str
    min_size: int = None
    max_size: int = None
    regex: re.Pattern = None

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_size: int = 0,
        max_size: int = max_int,
        regex: str = None,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        if min_size < 0:
            raise FieldValueError('String min_size cloud not be a negative number.')
        super().__init__(default=default, regex=regex, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Validates a string for its size and regex if they are specified.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional): The expected type of the attribute. If provided, the value is checked to ensure it is of this type. Defaults to None.

        Raises:
            FieldValueError: If the value does not match a regex, is not the
            correct type, or is an invalid length.
        """
        # First let all default checks be done.
        super().validate(attr_name, value, attr_type)

        if self.regex and value and value is not Undefined and not self.regex.match(value):
            raise FieldValueError(
                f"The value '{value}' for field '{attr_name}' must match with this regex: {self.regex.pattern}."
            )

        # Then we check for the size.
        try:
            if value is not None and value is not Undefined:
                _min_max_validate(min_value=self.min_size, max_value=self.max_size, value=len(value), attr_name=attr_name)
        except FieldValueError as error:
            raise FieldValueError(
                f"The length '{len(value)}' for attribute '{attr_name}' must be between '{self.min_size}' and '{self.max_size}'."
            ) from error


###############################################################################
#   RegexField Class Implementation
###############################################################################
class RegexField(Field):
    attr_type: re.Pattern = re.Pattern

    def __new__(cls, *args, **kwargs) -> re.Pattern:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        choices: set = None,
        **kwargs
    ) -> None:
        super().__init__(default=default, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, choices=choices, **kwargs)


    def clean_value(self, value: Any) -> Any:
        """
        Clean the value by compiling it into a regular expression pattern.

        This method takes a value and compiles it into a regular expression pattern if it is a string.
        If the value is equal to the default value, it updates the default value with the compiled pattern.
        It then calls the parent class's clean_value method to perform any additional cleaning.

        Args:
            value (Any): The value to be cleaned.

        Returns:
            Any: The cleaned value.

        Example:
            >>> field = RegexField(default='[0-9]+')
            >>> field.clean_value('123')
            re.compile('[0-9]+')
        """
        if isinstance(value, str):
            value = re.compile(value)
        return super().clean_value(value)


###############################################################################
#   TupleField Class Implementation
###############################################################################
class TupleField(Field):
    attr_type: tuple = tuple

    def __new__(cls, *args, **kwargs) -> tuple:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)


###############################################################################
#   URLField Class Implementation
###############################################################################
class URLField(Field):
    attr_type: str = str
    supported_schemes: set = {'http', 'https', 'ftp', 'ftps', 'git'}

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type | UnionType = None) -> None:
        """
        Validates a URL string to ensure it is a valid URL.
        We check the protocol HTTP, HTTPS, FTP, FTPS, the domain format and size and the port number.

        Args:
            attr_name (str): The name of the attribute being validated. This is used for identification in error messages.
            value (Any): The value to be validated for the specified attribute.
            attr_type (type | UnionType, optional): The expected type of the attribute. If provided, the value is checked to ensure it is of this type. Defaults to None.

        Raises:
            FieldValueError: If the url is invalid.
        """
        if isinstance(value, str):
            # https://github.com/django/django/blob/main/django/core/validators.py
            ul = "\u00a1-\uffff"  # Unicode letters range (must not be a raw string).

            # IP patterns
            ipv4_re = (
                r"(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)"
                r"(?:\.(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)){3}"
            )
            ipv6_re = r"\[[0-9a-f:.]+\]"  # (simple regex, validated later)

            # Host patterns
            hostname_re = (
                r"[a-z" + ul + r"0-9](?:[a-z" + ul + r"0-9-]{0,61}[a-z" + ul + r"0-9])?"
            )

            # Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
            domain_re = r"(?:\.(?!-)[a-z" + ul + r"0-9-]{1,63}(?<!-))*"
            tld_re = (
                r"\."  # dot
                r"(?!-)"  # can't start with a dash
                r"(?:[a-z" + ul + "-]{2,63}"  # domain label
                r"|xn--[a-z0-9]{1,59})"  # or punycode label
                r"(?<!-)"  # can't end with a dash
                r"\.?"  # may have a trailing dot
            )
            host_re = "(" + hostname_re + domain_re + tld_re + "|localhost)"

            regex = re.compile(
                r"^(?:[a-z0-9.+-]*)://"  # scheme is validated separately
                r"(?:[^\s:@/]+(?::[^\s:@/]*)?@)?"  # user:pass authentication
                r"(?:" + ipv4_re + "|" + ipv6_re + "|" + host_re + ")"
                r"(?::[0-9]{1,5})?"  # port
                r"(?:[/?#][^\s]*)?"  # resource path
                r"\Z",
                re.IGNORECASE,
            )
            if not regex.match(value):
                raise FieldValueError(f'Key {attr_name} must be an URL.')

            # The scheme needs a separate check
            try:
                result = urlparse(value)
                if result.scheme not in self.supported_schemes:
                    raise FieldValueError(f'Key {attr_name} must be an URL.')

            except ValueError:
                raise FieldValueError(f'Key {attr_name} must be an URL.') from ValueError

        return super().validate(attr_name, value, attr_type)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources import MarketData
from everysk.core.unittests import TestCase, mock


###############################################################################
#   Market Data TestCase Implementation
###############################################################################
class APIMarketDataTestCase(TestCase):

    def test_refresh(self):
        market_data = MarketData(retrieve_params={'params1': 'value1', 'params2': 'value2'}, params={'param_key1': 'param', 'param_key2': 'params3'})
        self.assertEqual(market_data.refresh(), market_data)

    def test_class_name(self):
        self.assertEqual(MarketData.class_name(), 'market_data')

    def test_class_url(self):
        self.assertEqual(MarketData.class_url(), '/market_data')

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_market_data__call_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()

        expected_response = {'success': True}
        mock_api_requestor.post.return_value = expected_response
        mock_create_api_requestor.return_value = mock_api_requestor
        response = MarketData._MarketData__call_method('test_method', key='value')

        mock_create_api_requestor.assert_called_once()
        mock_api_requestor.post.assert_called_once_with('/market_data/test_method', {'key': 'value'})
        self.assertEqual(response, expected_response)

    def test_symbols_search_method(self):
        with mock.patch('everysk.api.api_resources.market_data.MarketData._MarketData__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = MarketData.symbolsSearch(key='value')

        mock_call_method.assert_called_once_with('symbols_search', key='value')
        self.assertEqual(result, 'expected result')

    def test_symbols_check_method(self):
        with mock.patch('everysk.api.api_resources.market_data.MarketData._MarketData__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = MarketData.symbolsCheck(key='value')

        mock_call_method.assert_called_once_with('symbols_check', key='value')
        self.assertEqual(result, 'expected result')

    def test_symbols_price_method(self):
        with mock.patch('everysk.api.api_resources.market_data.MarketData._MarketData__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = MarketData.symbolsPrice(key='value')

        mock_call_method.assert_called_once_with('symbols_price', key='value')
        self.assertEqual(result, 'expected result')

    def test_symbols_realtime_price_method(self):
        with mock.patch('everysk.api.api_resources.market_data.MarketData._MarketData__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = MarketData.symbolsRealtimePrice(key='value')

        mock_call_method.assert_called_once_with('symbols_real_time_prices', key='value')
        self.assertEqual(result, 'expected result')

    def test_symbols_historical_method(self):
        with mock.patch('everysk.api.api_resources.market_data.MarketData._MarketData__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = MarketData.symbolsHistorical(key='value')

        mock_call_method.assert_called_once_with('symbols_historical', key='value')
        self.assertEqual(result, 'expected result')



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import re
from collections.abc import Iterator
from sys import maxsize as max_int
from typing import Any
from zoneinfo import ZoneInfo

from everysk.core import fields
from everysk.core.datetime.date import date, Date
from everysk.core.datetime.datetime import datetime, DateTime
from everysk.core.exceptions import ReadonlyError, RequiredError, FieldValueError
from everysk.core.object import BaseObject, BaseDict
from everysk.core.unittests import TestCase


class FieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.Field()
        self.assertEqual(field.attr_type, Any)

    def test_vscode_autocomplete(self):
        for cls in fields.Field.__subclasses__():
            self.assertEqual(cls.__new__.__annotations__['return'], cls.attr_type)

    def test_choices_empty(self):
        field = fields.Field(choices=set())
        self.assertEqual(field.choices, set())

    def test_choices_default(self):
        choices = {'a', 'b'}
        with self.assertRaisesRegex(FieldValueError, f"The default value 'c' must be in this list {choices}."):
            fields.Field(default='c', choices=choices)

    def test_choices(self):
        field = fields.Field(choices={'a', 'b'})
        self.assertEqual(field.choices, {'a', 'b'})
        with self.assertRaisesRegex(FieldValueError, f"The value 'c' for field 'banana' must be in this list {field.choices}."):
            field.validate(attr_name='banana', value='c', attr_type=str)


class BoolFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.BoolField()
        self.assertEqual(field.attr_type, bool)

    def test_clean_value_true(self):
        field = fields.BoolField()
        self.assertTrue(field.clean_value(1))
        self.assertTrue(field.clean_value(True))
        self.assertTrue(field.clean_value('y'))
        self.assertTrue(field.clean_value('yes'))
        self.assertTrue(field.clean_value('on'))

    def test_clean_value_false(self):
        field = fields.BoolField()
        self.assertFalse(field.clean_value(0))
        self.assertFalse(field.clean_value(False))
        self.assertFalse(field.clean_value('n'))
        self.assertFalse(field.clean_value('no'))
        self.assertFalse(field.clean_value('off'))

    def test_clean_value_error(self):
        field = fields.BoolField()
        with self.assertRaisesRegex(ValueError, "Invalid truth value 'banana'"):
            field.clean_value('BANANA')

    def test_default_value_readonly(self):
        with self.assertRaisesRegex(RequiredError, 'If field is readonly, then default value is required.'):
            fields.BoolField(default=None, readonly=True)

        fields.BoolField(default=True, readonly=True)
        fields.BoolField(default=False, readonly=True)


class ChoiceFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.ChoiceField(default='1', choices=['1', '2'])
        self.assertEqual(field.attr_type, str)

    def test_validate(self):
        field = fields.ChoiceField(default='1', choices=['1', '2'])
        field.validate(attr_name='test', value='2', attr_type=str)
        self.assertRaises(FieldValueError, field.validate, attr_name='test', value='3', attr_type=str)

    def test_undefined(self):
        # Allow Undefined as a value outside of the list
        field = fields.ChoiceField(default=Undefined, choices=['1', '2'])
        field.validate(attr_name='test', value=Undefined, attr_type=str)

    def test_choices_list(self):
        field = fields.ChoiceField(default='', choices=None)
        self.assertListEqual(field.choices, [])


class DateFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.DateField()
        self.assertEqual(field.attr_type, Date)

    def test_clean_value(self):
        field = fields.DateField()
        self.assertEqual(
            field.clean_value('2022-09-12'),
            Date(year=2022, month=9, day=12)
        )

        self.assertEqual(
            field.clean_value('20220912'),
            Date(year=2022, month=9, day=12)
        )

    def test_validate(self):
        field = fields.DateField()
        field.validate(attr_name='test', value=Date.fromisoformat('2023-01-01'))
        field.validate(attr_name='test', value=date.fromisoformat('2023-01-01'))# pylint: disable=no-member


class DateTimeFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.DateTimeField()
        self.assertEqual(field.attr_type, DateTime)

    def test_clean_value_str(self):
        field = fields.DateTimeField()
        self.assertEqual(
            field.clean_value('2022-09-12T13:00:00+00:00'),
            DateTime(year=2022, month=9, day=12, hour=13, tzinfo=ZoneInfo('UTC'))
        )
        self.assertEqual(
            field.clean_value('2022-09-12 13:00:00+00:00'),
            DateTime(year=2022, month=9, day=12, hour=13, tzinfo=ZoneInfo('UTC'))
        )
        self.assertEqual(
            field.clean_value('2022-09-14'),
            DateTime(year=2022, month=9, day=14, hour=0, tzinfo=ZoneInfo('UTC'))
        )
        self.assertEqual(
            field.clean_value('20220914'),
            DateTime(year=2022, month=9, day=14, hour=0, tzinfo=ZoneInfo('UTC'))
        )
        self.assertEqual(
            field.clean_value('20220914 13:00:00'),
            DateTime(year=2022, month=9, day=14, hour=13, tzinfo=ZoneInfo('UTC'))
        )

    def test_clean_value_str_date(self):
        field = fields.DateTimeField()
        self.assertEqual(
            field.clean_value('2022-09-12'),
            DateTime(year=2022, month=9, day=12, tzinfo=ZoneInfo('UTC'))
        )

    def test_clean_value_date(self):
        field = fields.DateTimeField()
        self.assertEqual(
            field.clean_value(Date(2022, 9, 12)),
            DateTime(year=2022, month=9, day=12, tzinfo=ZoneInfo('UTC'))
        )

    def test_validate(self):
        field = fields.DateTimeField()
        field.validate(attr_name='test', value=DateTime.fromisoformat('2022-09-12T13:00:00+00:00'))
        field.validate(attr_name='test', value=datetime.fromisoformat('2022-09-12T13:00:00+00:00')) # pylint: disable=no-member


class DictFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.DictField()
        self.assertEqual(field.attr_type, dict | BaseDict)

    def test_readonly(self):
        field = fields.DictField(default={'a': 1}, readonly=True)
        self.assertRaises(ReadonlyError, field.default.__setitem__, 'a', 2)
        self.assertRaises(ReadonlyError, field.default.__delitem__, 'a')
        self.assertRaises(ReadonlyError, field.default.pop)
        self.assertRaises(ReadonlyError, field.default.popitem, 'a')
        self.assertRaises(ReadonlyError, field.default.clear)
        self.assertRaises(ReadonlyError, field.default.update, {'a': 2})
        self.assertRaises(ReadonlyError, field.default.setdefault, 'b', 2)
        self.assertDictEqual(field.default, {'a': 1})

    def test_int_keys(self):
        class DictObjectIntKey(BaseObject):
            field = fields.DictField(default={1: 0})

        obj = DictObjectIntKey()
        self.assertEqual(obj.field[1], 0)
        obj.field = {1: 1}
        self.assertEqual(obj.field[1], 1)
        obj.field[1] = 2
        self.assertEqual(obj.field[1], 2)

    def test_int_keys_readonly(self):
        class DictObjectIntKey(BaseObject):
            field = fields.DictField(default={1: 0}, readonly=True)

        obj = DictObjectIntKey()
        self.assertEqual(obj.field[1], 0)

    def test_base_dict(self):
        field = fields.DictField(default=BaseDict(a=1))
        field.default = BaseDict(a=1)
        self.assertEqual(field.default, BaseDict(a=1))


class EmailFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.EmailField()
        self.assertEqual(field.attr_type, str)

    def test_valid_email(self):
        field = fields.EmailField()
        field.validate(attr_name='test', value='fscariott@everysk.com', attr_type=field.attr_type)

    def test_invalid_email(self):
        field = fields.EmailField()
        with self.assertRaisesRegex(FieldValueError, 'Key test must be an e-mail.'):
            field.validate(attr_name='test', value='fscariott.everysk.com', attr_type=field.attr_type)

    def test_invalid_email_with_more_at(self):
        field = fields.EmailField()
        with self.assertRaisesRegex(FieldValueError, 'Key test must be an e-mail.'):
            field.validate(attr_name='test', value='fscariott@everysk@com', attr_type=field.attr_type)


class FloatFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.FloatField()
        self.assertEqual(field.attr_type, float)

    def test_clean_value_str(self):
        field = fields.FloatField()
        self.assertEqual(field.clean_value('1.1'), 1.1)

    def test_clean_value_int(self):
        field = fields.FloatField()
        self.assertEqual(field.clean_value(2), 2.0)

    def test_min_size(self):
        field = fields.FloatField(min_size=1.5, max_size=2)
        self.assertRaisesRegex(
            FieldValueError,
            "The value '1' for field 'banana' must be between 1.5 and 2.",
            field.validate,
            attr_name='banana',
            value=1
        )

    def test_max_size(self):
        field = fields.FloatField(min_size=1.5, max_size=2)
        self.assertRaisesRegex(
            FieldValueError,
            "The value '2.5' for field 'banana' must be between 1.5 and 2.",
            field.validate,
            attr_name='banana',
            value=2.5
        )

    def test_none(self):
        field = fields.FloatField(min_size=1.5, max_size=2)
        self.assertIsNone(field.validate(attr_name='banana', value=None))


class IntFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.IntField()
        self.assertEqual(field.attr_type, int)

    def test_clean_value(self):
        field = fields.IntField()
        self.assertEqual(field.clean_value('1'), 1)

    def test_min_size(self):
        field = fields.IntField(min_size=1, max_size=2)
        self.assertRaisesRegex(
            FieldValueError,
            "The value '0' for field 'banana' must be between 1 and 2.",
            field.validate,
            attr_name='banana',
            value=0
        )

    def test_max_size(self):
        field = fields.IntField(min_size=1, max_size=2)
        self.assertRaisesRegex(
            FieldValueError,
            "The value '3' for field 'banana' must be between 1 and 2.",
            field.validate,
            attr_name='banana',
            value=3
        )

    def test_none(self):
        field = fields.IntField(min_size=1.5, max_size=2)
        self.assertIsNone(field.validate(attr_name='banana', value=None))


class IteratorFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.IteratorField()
        self.assertEqual(field.attr_type, Iterator)

    def test_clean_value(self):
        field = fields.IteratorField()
        self.assertIsInstance(field.clean_value('aaa'), Iterator)
        self.assertIsInstance(field.clean_value([0, 1]), Iterator)


class ListFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.ListField()
        self.assertEqual(field.attr_type, list)

    def test_clean_value(self):
        field = fields.ListField()
        self.assertEqual(field.clean_value('aaa'), ['aaa'])
        self.assertEqual(field.clean_value([0, 1]), [0, 1])
        self.assertEqual(field.clean_value(1), 1)
        self.assertEqual(field.clean_value({'a': 1}), {'a': 1})

    def test_max_size_default(self):
        # https://stackoverflow.com/questions/855191/how-big-can-a-python-list-get#comment112727918_15739630
        field = fields.ListField()
        self.assertEqual(field.max_size, max_int / 8)

    def test_max_size(self):
        field = fields.ListField(min_size=1, max_size=2)
        self.assertRaisesRegex(
            FieldValueError,
            "The attribute 'banana' is not within the specified list range. min_size: 1 max_size: 2",
            field.validate,
            attr_name='banana',
            value=[1, 2, 3]
        )

    def test_min_size(self):
        field = fields.ListField(min_size=2, max_size=3)
        self.assertRaisesRegex(
            FieldValueError,
            "The attribute 'banana' is not within the specified list range. min_size: 2 max_size: 3",
            field.validate,
            attr_name='banana',
            value=[1]
        )

    def test_min_size_negative(self):
        self.assertRaisesRegex(
            FieldValueError,
            'List min_size cloud not be a negative number.',
            fields.ListField,
            min_size=-1,
            max_size=3
        )

    def test_validate(self):
        field = fields.ListField()
        self.assertRaisesRegex(
            FieldValueError,
            "The 'banana' value must be a list.",
            field.validate,
            attr_name='banana',
            value=1234
        )

    def test_default_choices(self):
        fields.ListField(default=['a'], choices=['a', 'b'])

    def test_default_choices_raise_error(self):
        with self.assertRaises(FieldValueError) as error:
            fields.ListField(default=['c'], choices=['a', 'b'])
        self.assertEqual(error.exception.msg, "The default value '['c']' must be in this list ['a', 'b'].")

    def test_choices(self):
        field = fields.ListField(choices=['a', 'b'])
        field.validate(attr_name='banana', value=['a'])

    def test_choices_raise_error(self):
        field = fields.ListField(choices=['a', 'b'])
        with self.assertRaises(FieldValueError) as error:
            field.validate(attr_name='banana', value=['c'])
        self.assertEqual(error.exception.msg, "The value '['c']' for field 'banana' must be in this list ['a', 'b'].")

    def test_readonly(self):
        field = fields.ListField(default=[9, 1, 2, 3], readonly=True)
        self.assertRaises(ReadonlyError, field.default.__setitem__, 0, 1)
        self.assertRaises(ReadonlyError, field.default.__delitem__, 0)
        self.assertRaises(ReadonlyError, field.default.append, 4)
        self.assertRaises(ReadonlyError, field.default.clear)
        self.assertRaises(ReadonlyError, field.default.extend, [5, 6])
        self.assertRaises(ReadonlyError, field.default.insert, 0, 7)
        self.assertRaises(ReadonlyError, field.default.pop)
        self.assertRaises(ReadonlyError, field.default.remove, 0)
        self.assertRaises(ReadonlyError, field.default.reverse)
        self.assertRaises(ReadonlyError, field.default.sort)
        self.assertListEqual(field.default, [9, 1, 2, 3])

    def test_string_list(self):
        field = fields.ListField(default='a, b,   c,d, e   ')
        self.assertEqual(field.default, ['a', 'b', 'c', 'd', 'e'])
        field = fields.ListField(default='a; b;   c;d; e   ', separator=';')
        self.assertEqual(field.default, ['a', 'b', 'c', 'd', 'e'])


class SetFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.SetField()
        self.assertEqual(field.attr_type, set)

    def test_clean_value(self):
        field = fields.SetField()
        self.assertEqual(field.clean_value({0, 1}), {0, 1})
        self.assertEqual(field.clean_value(1), 1)
        self.assertEqual(field.clean_value({'a': 1}), {'a': 1})

    def test_max_size_default(self):
        # https://stackoverflow.com/questions/855191/how-big-can-a-python-list-get#comment112727918_15739630
        field = fields.SetField()
        self.assertEqual(field.max_size, max_int / 8)

    def test_max_size(self):
        field = fields.SetField(min_size=1, max_size=2)
        self.assertRaisesRegex(
            FieldValueError,
            "The attribute 'banana' is not within the specified set range. min_size: 1 max_size: 2",
            field.validate,
            attr_name='banana',
            value={1, 2, 3}
        )

    def test_min_size(self):
        field = fields.SetField(min_size=2, max_size=3)
        self.assertRaisesRegex(
            FieldValueError,
            "The attribute 'banana' is not within the specified set range. min_size: 2 max_size: 3",
            field.validate,
            attr_name='banana',
            value={1}
        )

    def test_min_size_negative(self):
        self.assertRaisesRegex(
            FieldValueError,
            'Set min_size cloud not be a negative number.',
            fields.SetField,
            min_size=-1,
            max_size=3
        )

    def test_validate(self):
        field = fields.SetField()
        self.assertRaisesRegex(
            FieldValueError,
            "Key banana must be <class 'set'>.",
            field.validate,
            attr_name='banana',
            value=1234
        )

    def test_readonly(self):
        field = fields.SetField(default={9, 1, 2, 3}, readonly=True)
        self.assertRaises(ReadonlyError, field.default.__setitem__, 0, 1)
        self.assertRaises(ReadonlyError, field.default.__delitem__, 0)
        self.assertRaises(ReadonlyError, field.default.add, 4)
        self.assertRaises(ReadonlyError, field.default.clear)
        self.assertRaises(ReadonlyError, field.default.discard, 1)
        self.assertRaises(ReadonlyError, field.default.pop)
        self.assertRaises(ReadonlyError, field.default.remove, 1)
        self.assertRaises(ReadonlyError, field.default.update, {1, 4})
        self.assertSetEqual(field.default, {9, 1, 2, 3})

    def test_default_choices(self):
        fields.SetField(default={'a'}, choices={'a', 'b'})

    def test_default_choices_raise_error(self):
        with self.assertRaises(FieldValueError) as error:
            fields.SetField(default={'c'}, choices={'a'})
        self.assertEqual(error.exception.msg, "The default value '{'c'}' must be in this list {'a'}.")

    def test_choices(self):
        field = fields.SetField(choices={'a', 'b'})
        field.validate(attr_name='banana', value={'a'})

    def test_choices_raise_error(self):
        field = fields.SetField(choices={'a'})
        with self.assertRaises(FieldValueError) as error:
            field.validate(attr_name='banana', value={'c'})
        self.assertEqual(error.exception.msg, "The value '{'c'}' for field 'banana' must be in this list {'a'}.")


class StrFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.StrField()
        self.assertEqual(field.attr_type, str)

    def test_max_size(self):
        field = fields.StrField(min_size=1, max_size=2)
        self.assertRaisesRegex(
            FieldValueError,
            "The length '3' for attribute 'banana' must be between '1' and '2'.",
            field.validate,
            attr_name='banana',
            value='abc'
        )

    def test_min_size(self):
        field = fields.StrField(min_size=2, max_size=3)
        self.assertRaisesRegex(
            FieldValueError,
            "The length '1' for attribute 'banana' must be between '2' and '3'.",
            field.validate,
            attr_name='banana',
            value='a'
        )

    def test_min_size_negative(self):
        self.assertRaisesRegex(
            FieldValueError,
            'String min_size cloud not be a negative number.',
            fields.StrField,
            min_size=-1,
            max_size=3
        )

    def test_validate(self):
        field = fields.StrField()
        self.assertRaisesRegex(
            FieldValueError,
            "Key banana must be <class 'str'>.",
            field.validate,
            attr_name='banana',
            value=1234
        )

        field.regex = re.compile('[a-z]')
        with self.assertRaises(FieldValueError) as e:
            field.validate(attr_name='banana', value='1234')
        self.assertEqual("The value '1234' for field 'banana' must match with this regex: [a-z].", e.exception.msg)


class TupleFieldTestCase(TestCase):

    def test_attr_type(self):
        field = fields.TupleField()
        self.assertEqual(field.attr_type, tuple)


class FieldUndefinedClass(BaseObject):
    f01 = fields.BoolField(default=Undefined)
    f02 = fields.ChoiceField(default=Undefined, choices=[Undefined, None])
    f03 = fields.DateField(default=Undefined)
    f04 = fields.DateTimeField(default=Undefined)
    f05 = fields.FloatField(default=Undefined)
    f06 = fields.IntField(default=Undefined)
    f07 = fields.IteratorField(default=Undefined)
    f08 = fields.ListField(default=Undefined)
    f09 = fields.StrField(default=Undefined)
    f10 = fields.TupleField(default=Undefined)


class FieldUndefinedTestCase(TestCase):

    def test_bool_field(self):
        obj = FieldUndefinedClass(f01=Undefined)
        obj.f01 = Undefined

    def test_choice_field(self):
        obj = FieldUndefinedClass(f02=Undefined)
        obj.f02 = Undefined

    def test_date_field(self):
        obj = FieldUndefinedClass(f03=Undefined)
        obj.f03 = Undefined

    def test_datetime_field(self):
        obj = FieldUndefinedClass(f04=Undefined)
        obj.f04 = Undefined

    def test_float_field(self):
        obj = FieldUndefinedClass(f05=Undefined)
        obj.f05 = Undefined

    def test_int_field(self):
        obj = FieldUndefinedClass(f06=Undefined)
        obj.f06 = Undefined

    def test_iterator_field(self):
        obj = FieldUndefinedClass(f07=Undefined)
        obj.f07 = Undefined

    def test_list_field(self):
        obj = FieldUndefinedClass(f08=Undefined)
        obj.f08 = Undefined

    def test_str_field(self):
        obj = FieldUndefinedClass(f09=Undefined)
        obj.f09 = Undefined

    def test_tuple_field(self):
        obj = FieldUndefinedClass(f10=Undefined)
        obj.f10 = Undefined


## https://everysk.atlassian.net/browse/COD-3457
class BaseObjectPropertyInit(BaseObject):
    p1 = fields.IntField()

    @property
    def p2(self) -> int:
        return self.p1

    @p2.setter
    def p2(self, value: int) -> None:
        self.p1 = value

class BaseDictPropertyInit(BaseDict):
    p1 = fields.IntField()

    @property
    def p2(self) -> int:
        return self.p1

    @p2.setter
    def p2(self, value: int) -> None:
        self.p1 = value

class ObjectInitPropertyTestCase(TestCase):

    def test_base_object_normal_attribute(self):
        obj = BaseObjectPropertyInit(p1=1)
        self.assertEqual(obj.p1, 1)
        self.assertEqual(obj.p2, 1)

    def test_base_object_property_attribute(self):
        obj = BaseObjectPropertyInit(p2=2)
        self.assertEqual(obj.p1, 2)
        self.assertEqual(obj.p2, 2)

    def test_base_dict_normal_attribute(self):
        obj = BaseDictPropertyInit(p1=1)
        self.assertEqual(obj.p1, 1)
        self.assertEqual(obj.p2, 1)

    def test_base_dict_property_attribute(self):
        obj = BaseDictPropertyInit(p2=2)
        self.assertEqual(obj.p1, 2)
        self.assertEqual(obj.p2, 2)


# https://everysk.atlassian.net/browse/COD-3770
class COD3770(BaseObject):
    var = fields.DateField(default='2024-01-01')


class COD3770TestCase(TestCase):

    def setUp(self) -> None:
        self.date = Date(2024, 1, 1)

    def test_field_default(self):
        obj = fields.DateField(default='2024-01-01')
        self.assertEqual(obj.default, self.date)

    def test_class_var(self):
        self.assertEqual(COD3770.var, self.date)

    def test_class_var_attribution(self):
        # We create a specific class to not change other tests
        class VarTest(BaseObject):
            var = fields.DateField(default='2024-01-01')

        self.assertEqual(VarTest.var, self.date)
        VarTest.var = '2024-01-02'
        self.assertEqual(VarTest.var, Date(2024, 1, 2))

    def test_instance_var_default(self):
        obj = COD3770()
        self.assertEqual(obj.var, self.date)

    def test_instance_var_attribution(self):
        obj = COD3770()
        obj.var = '2024-01-02'
        self.assertEqual(obj.var, Date(2024, 1, 2))


class URLFieldTestCase(TestCase):

    class ObjectClass(BaseObject):
        url = fields.URLField()

    class DictClass(BaseDict):
        url = fields.URLField()

    def test_http_url(self):
        url = 'http://exampe.com/path?var=1&var=2'
        obj = self.ObjectClass()
        dct = self.DictClass()
        obj.url = url
        dct['url'] = url
        self.assertEqual(obj.url, dct['url'])

    def test_https_url(self):
        url = 'https://exampe.com/path?var=1&var=2'
        obj = self.ObjectClass()
        dct = self.DictClass()
        obj.url = url
        dct['url'] = url
        self.assertEqual(obj.url, dct['url'])

    def test_localhost_url(self):
        url = 'http://localhost:8080/path?var=1&var=2'
        obj = self.ObjectClass()
        dct = self.DictClass()
        obj.url = url
        dct['url'] = url
        self.assertEqual(obj.url, dct['url'])

    def test_not_valid_scheme(self):
        url = 'ttp://localhost:8080/path?var=1&var=2'
        obj = self.ObjectClass()
        dct = self.DictClass()
        with self.assertRaisesRegex(FieldValueError, 'Key url must be an URL.'):
            obj.url = url
        with self.assertRaisesRegex(FieldValueError, 'Key url must be an URL.'):
            dct['url'] = url

    def test_not_valid_domain(self):
        url = 'http://localhost.host:/path?var=1&var=2'
        obj = self.ObjectClass()
        dct = self.DictClass()
        with self.assertRaisesRegex(FieldValueError, 'Key url must be an URL.'):
            obj.url = url
        with self.assertRaisesRegex(FieldValueError, 'Key url must be an URL.'):
            dct['url'] = url

    def test_not_valid_query(self):
        url = 'http://localhost:8080/path?var=1 var=2'
        obj = self.ObjectClass()
        dct = self.DictClass()
        with self.assertRaisesRegex(FieldValueError, 'Key url must be an URL.'):
            obj.url = url
        with self.assertRaisesRegex(FieldValueError, 'Key url must be an URL.'):
            dct['url'] = url



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['Logger', 'LoggerManager']

import json
import logging
import traceback as python_traceback
import sys
from contextlib import AbstractContextManager
from contextvars import ContextVar
from datetime import datetime as DateTime
from types import TracebackType
from typing import Any
from threading import Thread
from zoneinfo import ZoneInfo

from everysk.config import settings


DEFAULT_APP_SERVER = settings.LOGGING_APP_SERVER
DEFAULT_TRACE_PARTS = {'version': '', 'trace_id': '', 'span_id': '', 'trace_sample': False}
HEADER_TRACEPARENT = 'traceparent'
HEADER_X_CLOUD_TRACE_CONTEXT = 'x-cloud-trace-context'
LOGGER_MANAGER_CONTEXT_VAR_NAME = 'everysk-lib-log-extra-context-var'


###############################################################################
#   Private functions Implementation
###############################################################################
def _default(obj: Any) -> str | None:
    """
    This function is used to convert the object to a string inside the json.dumps.

    Args:
        obj (Any): The object to be converted to a string.
    """
    if isinstance(obj, bytes):
        return obj.decode(json.detect_encoding(obj))

    return None

def _get_gcp_headers(headers: dict = None) -> dict:
    """
    Get the headers to be added to the log.
    The order is if the headers are sent in the log function, set in context or get from the default function.

    Args:
        headers (dict, optional): The headers generated outside the log. Defaults to None.

    Returns:
        dict: Only the headers that are in the list HEADER_TRACEPARENT and HEADER_X_CLOUD_TRACE_CONTEXT.
    """
    # pylint: disable=protected-access
    if not headers:
        headers = LoggerManager._extra.get().get('http_headers', {})

    return {key: value for key, value in headers.items() if key in (HEADER_TRACEPARENT, HEADER_X_CLOUD_TRACE_CONTEXT)}

def _get_trace_data(headers: dict) -> dict:
    """
    Get the trace_id, span_id and trace_sample from the headers.
    https://cloud.google.com/trace/docs/trace-context

    Args:
        headers (dict): The headers dictionary.
    """
    trace_parts = DEFAULT_TRACE_PARTS.copy()
    trace = headers.get(HEADER_TRACEPARENT, '')
    if trace:
        _parse_traceparent(trace, trace_parts)
    else:
        trace = headers.get(HEADER_X_CLOUD_TRACE_CONTEXT, '')
        if trace:
            _parse_x_cloud_trace_context(trace, trace_parts)

    if trace_parts['trace_id']:
        trace_parts['trace_id'] = f"{settings.LOGGING_GOOGLE_CLOUD_TRACE_ID}/{trace_parts['trace_id']}"

    return trace_parts

def _get_traceback() -> str:
    """
    Get the traceback of the current exception.
    """
    result = python_traceback.format_exc()
    # When there is no traceback the result is 'NoneType: None\n'
    # Like this is the most common case we check for equality and return an empty string
    if result == 'NoneType: None\n':
        return ''

    return result

def _parse_traceparent(traceparent: str, trace_parts: dict) -> None:
    """
    Parse the traceparent header and return a dictionary with the version, trace_id, span_id and trace_flags.
    Header example -> traceparent: "00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01"
    https://www.w3.org/TR/trace-context/
    From W3C documentation: traceparent: "<version>-<trace-id>-<parent-id>-<trace-flags>"

    Args:
        traceparent (str): The traceparent header value.
    """
    try:
        trace_parts['version'], trace_parts['trace_id'], trace_parts['span_id'], trace_parts['trace_sample'] = traceparent.split('-')
        trace_parts['trace_sample'] = bool(int(trace_parts['trace_sample']))
    except ValueError:
        pass

def _parse_x_cloud_trace_context(trace_context: str, trace_parts: dict) -> None:
    """
    Parse the x-cloud-trace-context header and return a dictionary with the trace_id and span_id.
    Header example -> x-cloud-trace-context: "4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1"
    From Google documentation: X-Cloud-Trace-Context: TRACE_ID/SPAN_ID;o=OPTIONS

    Args:
        trace_context (str): The x-cloud-trace-context header value.
    """
    try:
        trace_parts['trace_id'], trace_parts['span_id'] = trace_context.split('/')
        trace_parts['span_id'], trace_parts['trace_sample'] = trace_parts['span_id'].split(';')
        trace_parts['trace_sample'] = trace_parts['trace_sample'].endswith('1') # pylint: disable=no-member
    except ValueError:
        pass


###############################################################################
#   Formatter Class Implementation
###############################################################################
class Formatter(logging.Formatter):

    def _get_default_dict(self, message: str, severity: str) -> dict:
        """
        Python logging default values.
        Severity levels: CRITICAL , ERROR , WARNING , INFO , DEBUG

        Args:
            message (str): The message to be logged.
            severity (str): The severity of the message.
        """
        return {'message': message, 'severity': severity}

    def _get_default_extra_dict(self, name: str, headers: dict, payload: dict, response: dict, traceback: str, labels: dict) -> dict:
        """
        Get the default extra data dictionary to be added to the log.
        Until now we only have the logName, traceback, http headers and http payload.

        Args:
            name (str): The name used to create the log.
            headers (dict): A dictionary with the HTTP headers.
            payload (dict): A dictionary with the HTTP payload.
        """
        return {
            'logName': name,
            'labels': labels,
            'traceback': traceback,
            'http': {
                'headers': headers,
                'payload': payload,
                'response': response,
            }
        }

    def _get_default_gcp_dict(self, headers: dict, filename: str, line: int, func_name: str) -> dict:
        """
        Default Google Cloud Platform dictionary to be added to the log.
        This dictionary has the trace_id, span_id, trace_sample used to chain the logs in the Google Cloud Logging.

        Args:
            headers (dict): A dictionary with the HTTP headers.
            filename (str): Filename where the log was placed.
            line (int): The line where the log was placed.
            func_name (str): The function name where the log was placed.
        """
        trace = _get_trace_data(headers)
        return {
            'logging.googleapis.com/trace': trace['trace_id'],
            'logging.googleapis.com/spanId': trace['span_id'],
            'logging.googleapis.com/trace_sampled': trace['trace_sample'],
            'logging.googleapis.com/sourceLocation': {
                'file': filename,
                'line': line,
                'function': func_name,
            },
        }

    def _get_result_dict(self, record: logging.LogRecord) -> dict:
        """
        Convert the log record to a dictionary.

        Args:
            record (logging.LogRecord): Record object with all the information about the log.
        """
        result = self._get_default_dict(message=record.getMessage(), severity=record.levelname)
        result.update(self._get_default_extra_dict(
            name=record.name,
            headers=getattr(record, 'http_headers', {}),
            payload=getattr(record, 'http_payload', {}),
            response=getattr(record, 'http_response', {}),
            traceback=getattr(record, 'traceback', ''),
            labels=getattr(record, 'labels', {})
        ))
        result.update(self._get_default_gcp_dict(
            headers=result['http']['headers'],
            filename=record.pathname,
            line=record.lineno,
            func_name=record.funcName)
        )
        return result

    def formatMessage(self, record: logging.LogRecord) -> str:
        """
        Format the message to be displayed in the terminal or Google Log Explorer.

        Args:
            record (logging.LogRecord): Record object with all the information about the log.
        """
        result = self._get_result_dict(record)
        return json.dumps(result, default=_default)

###############################################################################
#   LoggerManager Class Implementation
###############################################################################
class LoggerManager(AbstractContextManager):
    ## Private attributes
    _extra: ContextVar = ContextVar(LOGGER_MANAGER_CONTEXT_VAR_NAME, default={})
    _old_value: dict = None

    def __init__(
            self,
            http_headers: dict = None,
            http_payload: dict = None,
            http_response: dict = None,
            labels: dict = None,
            stacklevel: int = None,
            traceback: str = None,
        ) -> None:
        """
        Context class to create a context manager for the Logger object.
        This class is used to add extra information to the log.
        This manager together with everysk.core.threads will keep the context inside the thread.
        If you use python threads you need to pass the context inside the thread.

        Args:
            http_headers (dict, optional): The HTTP headers to be added to the log. Defaults to None.
            http_payload (dict, optional): The HTTP payload to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to None.
        """
        self.http_headers = http_headers
        self.http_payload = http_payload
        self.http_response = http_response
        self.labels = labels
        self.stacklevel = stacklevel
        self.traceback = traceback

        # Save this value to be used in the __exit__ method to restore the context
        self._old_value = self._extra.get().copy()

    def __enter__(self) -> 'LoggerManager':
        """
        Method to be executed when the context manager is created and always return self.
        This method is always executed even if "LoggerManager as some_var:" is not used.
        """
        # First we get the original context
        context: dict = self._extra.get()

        if self.http_headers is not None:
            context['http_headers'] = self.http_headers

        if self.http_payload is not None:
            context['http_payload'] = self.http_payload

        if self.http_response is not None:
            context['http_response'] = self.http_response

        if self.labels is not None:
            context['labels'] = self.labels

        if self.stacklevel is not None:
            context['stacklevel'] = self.stacklevel

        if self.traceback is not None:
            context['traceback'] = self.traceback

        # Finally we store the new context
        self._extra.set(context)

        return self

    def __exit__(self, __exc_type: type[BaseException] | None, __exc_value: BaseException | None, __traceback: TracebackType | None) -> bool | None:
        """
        https://docs.python.org/3/library/stdtypes.html#contextmanager.__exit__

        Returns:
            bool | None: If return is False any exception will be raised.
        """
        # Restore the context to the original value
        self._extra.set(self._old_value)

        return False

    @classmethod
    def reset(cls) -> None:
        """
        Reset the context to the default value.
        This method is used to avoid shared values between requests in server.endpoints module.
        """
        cls._extra = ContextVar(LOGGER_MANAGER_CONTEXT_VAR_NAME, default={})

###############################################################################
#   Logger Class Implementation
###############################################################################
class Logger:

    ## Private attributes
    _deprecated_hash: set[str] = set() # This needs to be initialized with an empty set because it will be a global list
    _default_stacklevel: int = 3 # 3 -> The place where the log.method was placed; 2 -> This file; 1 -> Python logger
    _log: logging.Logger = None
    _slack_timer: DateTime = None

    ## Public attributes
    name: str = None
    stacklevel: int = None

    ## Private methods
    def __init__(self, name: str, stacklevel: int = None) -> None:
        """
        Logger class used to send messages to STDOUT or Google CLoud Logging.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            name (str, optional): The name of the log. Defaults to "root".
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.

        Example:
            >>> from everysk.core.log import Logger
            >>> log = Logger(name='log-test')
            >>> log.debug('Test')
            2024-02-07 12:49:10,640 - DEBUG - {} - Test
        """
        if name == 'root':
            raise ValueError('The name of the log could not be "root".')
        self.name = name
        if stacklevel is not None and stacklevel > 0:
            self.stacklevel = stacklevel

        self._log = self._get_python_logger()

    def _get_extra_data(self, extra: dict, level: int = None) -> dict:
        """
        Get the extra data to be added to the log.
        We check if the extra was sent in the function, if not we check if it
        was set in the LoggerManager context, if not we check if we have a
        default function to get the data.
        We only set the payload

        Args:
            extra (dict): The extra data sent in the function.
            level (int): The level of the log.
        """
        http_headers = self._get_http_headers(extra.pop('http_headers', {}))
        if http_headers:
            extra['http_headers'] = http_headers

        http_payload = self._get_http_payload(extra.pop('http_payload', {}), level)
        if http_payload:
            extra['http_payload'] = http_payload

        http_response = self._get_http_response(extra.pop('http_response', {}))
        if http_response:
            extra['http_response'] = http_response

        # https://everysk.atlassian.net/browse/COD-6151
        # Because we use labels in the format for the normal log, this key always will be present
        labels = extra.pop('labels', {})
        if not labels:
            # If we don't receive labels as param in log functions we get from the context
            labels = LoggerManager._extra.get().get('labels', {}) # pylint: disable=protected-access

        # Labels are always present in the log even if it is an empty dictionary
        extra['labels'] = labels

        traceback = extra.pop('traceback', '')
        if not traceback:
            # If we don't receive traceback as param in log functions we get from the context
            # or from the result of traceback module
            traceback = LoggerManager._extra.get().get('traceback', '') or _get_traceback() # pylint: disable=protected-access
        if traceback:
            extra['traceback'] = traceback

        return extra

    def _get_http_headers(self, http_headers: dict) -> dict:
        """
        Get the http headers to be added to the log.
        The order is if the headers are sent in the log function, set in context or get from the default function.
        We only search for the payload if the level is ERROR or CRITICAL.

        Args:
            http_headers (dict): The HTTP headers sent in the log function.
            level (int): The level of the log.
        """
        return _get_gcp_headers(http_headers)

    def _get_http_payload(self, http_payload: dict, level: int = None) -> dict:
        """
        Get the http payload to be added to the log.
        The order is if the payload are sent in the log function, set in context or get from the default function.
        We only search for the payload if the level is ERROR or CRITICAL.

        Args:
            http_payload (dict): The HTTP payload sent in the log function.
            level (int): The level of the log.
        """
        if not http_payload and level in (logging.ERROR, logging.CRITICAL):
            http_payload = LoggerManager._extra.get().get('http_payload', {}) # pylint: disable=protected-access

        return http_payload

    def _get_http_response(self, http_response: dict) -> dict:
        """
        Get the http response to be added to the log.
        The order is if the response are sent in the log function, set in context or get from the default function.

        Args:
            http_response (dict): The HTTP response sent in the log function.
        """
        if not http_response:
            http_response = LoggerManager._extra.get().get('http_response', {}) # pylint: disable=protected-access

        return http_response

    def _get_python_logger(self) -> logging.Logger:
        """
        Method that creates/get the Python Logger object and attach the correct handler.
        The default handler will be the stdout.
        """
        if self._log is not None:
            return self._log

        # Create the log
        log = logging.getLogger(self.name)
        log.setLevel(logging.DEBUG)
        log.propagate = False # Don't pass message to others loggers

        # We should only have one handler per log name
        if not hasattr(log, 'handler'):
            log.handler = logging.StreamHandler(stream=sys.stdout)

            # Set the format that the message is displayed
            if settings.LOGGING_JSON:
                log.handler.setFormatter(Formatter())
            else:
                log.handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(labels)s - %(message)s'))

            # Add the handler inside the log
            log.addHandler(log.handler)

        # Set the level that the handler will be using.
        log.handler.setLevel(logging.DEBUG)

        return log

    def _get_stacklevel(self, stacklevel: int) -> int:
        """
        Get the stacklevel that was set in the function
        or in the LoggerManager context
        or set in the Logger object
        or use the default stacklevel.

        Args:
            stacklevel (int): The stacklevel to be used on the log.
        """
        if not stacklevel:
            stacklevel = LoggerManager._extra.get().get('stacklevel', {}) # pylint: disable=protected-access
            if not stacklevel:
                stacklevel = self.stacklevel or self._default_stacklevel

        return stacklevel

    def _send_to_log(self, level: int, msg: str, *args: tuple, extra: dict = None, stacklevel: int = None) -> None:
        """
        Send the message to the python logger using the correct level.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            level (int): The level of the message.
            msg (str): The message to log.
            *args (tuple): The arguments to be used on the message.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        stacklevel = self._get_stacklevel(stacklevel)
        extra = self._get_extra_data(extra or {}, level)
        self._log.log(level, msg, *args, extra=extra, stacklevel=stacklevel)

    def _show_deprecated(self, _id: str, show_once: bool) -> bool:
        """
        If show_once is False this always return True, otherwise
        checks if this _id is in the self._deprecated_hash set.

        Args:
            _id (str): String that will be stored to be checked later
            show_once (bool): Flag to store or not the id.
        """
        if show_once:
            if _id in self._deprecated_hash:
                return False

            self._deprecated_hash.add(_id)

        return True

    ## Public methods
    def critical(self, msg: str, *args: tuple, extra: dict = None, stacklevel: int = None)-> None:
        """
        Log a message with severity CRITICAL on this logger.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            msg (str): The message to log.
            *args (tuple): The arguments to be used on the message.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        self._send_to_log(logging.CRITICAL, msg, *args, extra=extra, stacklevel=stacklevel)

    def debug(self, msg: str, *args: tuple, extra: dict = None, stacklevel: int = None)-> None:
        """
        Log a message with severity DEBUG on this logger.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            msg (str): The message to log.
            *args (tuple): The arguments to be used on the message.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        self._send_to_log(logging.DEBUG, msg, *args, extra=extra, stacklevel=stacklevel)

    def deprecated(self, msg: str, *args, show_once: bool = True, extra: dict = None, stacklevel: int = None) -> None:
        """
        Shows a DeprecationWarning message with severity 'WARNING'.
        If show_once is True, then the message will be showed only once.

        Args:
            msg (str): The message that must be showed.
            show_once (bool, optional): If the message must be showed only once. Defaults to True.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        _id = hash(f'{msg}, {args}')
        if self._show_deprecated(_id=_id, show_once=show_once):
            msg = f'DeprecationWarning: {msg}'
            self.warning(msg, *args, extra=extra, stacklevel=stacklevel)

    def error(self, msg: str, *args: tuple, extra: dict = None, stacklevel: int = None)-> None:
        """
        Log a message with severity ERROR on this logger.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            msg (str): The message to log.
            *args (tuple): The arguments to be used on the message.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        self._send_to_log(logging.ERROR, msg, *args, extra=extra, stacklevel=stacklevel)

    def exception(self, msg: str, *args: tuple, extra: dict = None, stacklevel: int = None)-> None:
        """
        Log a message with severity ERROR on this logger.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            msg (str): The message to log.
            *args (tuple): The arguments to be used on the message.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        self.error(msg, *args, extra=extra, stacklevel=stacklevel)

    def info(self, msg: str, *args: tuple, extra: dict = None, stacklevel: int = None)-> None:
        """
        Log a message with severity INFO on this logger.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            msg (str): The message to log.
            *args (tuple): The arguments to be used on the message.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        self._send_to_log(logging.INFO, msg, *args, extra=extra, stacklevel=stacklevel)

    def _can_send_slack(self) -> bool:
        """
        Check if the slack message can be sent.
        We keep a track of the last message sent to avoid sending too many messages in less than 3 seconds.
        """
        now = DateTime.now(ZoneInfo('UTC')) # pylint: disable=no-member
        if self._slack_timer and (now - self._slack_timer).seconds < 2:
            return False

        self._slack_timer = now
        return True

    def slack(self, title: str, message: str, color: str, url: str = None, extra: dict = None, stacklevel: int = None) -> None:
        """
        Send a message to a Slack channel using Slack WebHooks.
        https://api.slack.com/messaging/webhooks
        The same message will be sent to the default log too:
            danger -> error
            success -> info
            warning -> warning

        Args:
            title (str): The title of the message.
            message (str): The body of the message.
            color (str): 'danger' | 'success' | 'warning'
            url (str, optional): The slack webhook url. Defaults to settings.SLACK_URL.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        if url is None:
            url = settings.SLACK_URL

        # We send the message only if url is set and we are in PROD and not in unittest
        if url and settings.PROFILE == 'PROD' and 'unittest' not in sys.modules:
            # We check if the last message was sent in less than 3 seconds to avoid sending too many messages
            if self._can_send_slack():
                # The import must be here to avoid circular import inside http module
                from everysk.core.slack import Slack # pylint: disable=import-outside-toplevel
                client = Slack(title=title, message=message, color=color, url=url)
                # This will send the message to Slack without block the request
                Thread(target=client.send).start()

        log_message = f'Slack message: {title} -> {message}'
        if color == 'danger':
            self.error(log_message, extra=extra, stacklevel=stacklevel)

        elif color == 'success':
            self.info(log_message, extra=extra, stacklevel=stacklevel)

        elif color == 'warning':
            self.warning(log_message, extra=extra, stacklevel=stacklevel)

    def warning(self, msg: str, *args: tuple, extra: dict = None, stacklevel: int = None)-> None:
        """
        Log a message with severity WARNING on this logger.
        Use stacklevel to show correctly the file and the line of the log, lvl 0 means the python
        logger, lvl 1 means this file, lvl 2 means the place where the log. was placed, if you use
        another file for the log object maybe you need to change to lvl 3.

        Args:
            msg (str): The message to log.
            *args (tuple): The arguments to be used on the message.
            extra (dict, optional): Extra information to be added to the log. Defaults to None.
            stacklevel (int, optional): The stacklevel to be used on the log. Defaults to 2.
        """
        self._send_to_log(logging.WARNING, msg, *args, extra=extra, stacklevel=stacklevel)
